import os
import glob
import ntpath
import sys
import pkg_resources
import traceback
import subprocess
import datetime
import numpy as np
import re
import twutils.plot as twplot
import matplotlib as mpl
import matplotlib.pyplot as plt
import PIL.Image

class term:
	ok = '\u2713'
	err = '\u2717'
	red = '\u001b[31m'
	green = '\u001b[32m'
	yellow = '\u001b[33m'
	blue = '\u001b[94m'
	cyan = '\u001b[96m'
	reset_color = '\u001b[39;49m'
	reset_all = '\u001b[0m'
	bold = '\u001b[1m'

def print_usage():
	print('USAGE:')
	print('  Recommend running from a scratch directory.')
	print('  twtest [FLAGS] [OPTIONS] --root <tw_root> --command <cmd>')
	print()
	print('FLAGS:')
	print('  --unit         unit tests')
	print('  --integration  integration tests')
	print('  --sea-trials   full scale examples')
	print('  -h, --help     display help')
	print('  -v, --version  display version')
	print()
	print('OPTIONS:')
	print('  --categories <cat1,cat2,...>  limit to explicit categories')
	print()
	print('ARGS:')
	print('  <tw_root>   path to turboWAVE root directory.')
	print('  <cmd>       command line to execute for each test.')
	print()
	print('EXAMPLES:')
	print('  twtest --categories hydro,pic --sea-trials --root ~/turboWAVE --command tw3d -n 4 -c 5')
	print('  twtest --integration --unit --root ~/turboWAVE --command mpirun -np 4 tw3d')

def print_version():
    print('twtest is provided by twutils, version '+pkg_resources.get_distribution('twutils').version)

def cleanup(wildcarded_path):
	cleanstr = glob.glob(wildcarded_path)
	for f in cleanstr:
		os.remove(f)

def gen_plot(img_file,plotter,slicing_spec,slices,dyn_range=0.0,val_range=(0.0,0.0),my_color_map='viridis',global_contrast=False):
	plt.figure(figsize=(7,5),dpi=75)
	if len(slices)==2:
		data_slice,plot_dict = plotter.falsecolor2d(slicing_spec,slices,dyn_range)
		if global_contrast:
			minval,maxval = plotter.get_global_minmax(dyn_range)
		else:
			minval = plot_dict['vmin']
			maxval = plot_dict['vmax']
		if val_range!=(0.0,0.0):
			minval = val_range[0]
			maxval = val_range[1]
		plt.imshow(data_slice,
			origin='lower',
			aspect='auto',
			extent=plot_dict['extent'],
			vmin=minval,
			vmax=maxval,
			cmap=my_color_map)
		bar = plt.colorbar()
		plt.xlabel(plot_dict['xlabel'],fontsize=18)
		plt.ylabel(plot_dict['ylabel'],fontsize=18)
		bar.set_label(plot_dict['blabel'],fontsize=18)
	if len(slices)==3:
		abcissa,ordinate,plot_dict = plotter.lineout(slicing_spec,slices,dyn_range)
		plt.plot(abcissa,ordinate)
		plt.xlabel(plot_dict['xlabel'],fontsize=18)
		plt.ylabel(plot_dict['ylabel'],fontsize=18)
	plt.tight_layout()
	plt.savefig(img_file)
	plt.close()

def gen_movie(mov_file,plotter,slicing_spec,slices,dyn_range=0.0,val_range=(0.0,0.0),my_color_map='viridis'):
	frames = plotter.dims4()[0]
	for i in range(frames):
		img_file = 'frame{:03d}.png'.format(i)
		c = slicing_spec.find('t')
		slices[c-2] = str(i)
		gen_plot(img_file,plotter,slicing_spec,tuple(map(int,slices)),dyn_range,val_range,my_color_map,True)
	images = []
	frameRateHz = 5
	img_files = sorted(glob.glob('frame*.png'))
	for f in img_files:
		images.append(PIL.Image.open(f))
	images[0].save(mov_file,save_all=True,append_images=images[1:],duration=int(1000/frameRateHz),loop=0)
	cleanup('frame*.png')

def gen_viz(primitive_file,fig_dict):
	'''Figure out whether a still image or movie was requested and call appropriate routine'''
	file_to_plot = fig_dict['data']
	slicing_spec = fig_dict['slicing_spec']
	slices = fig_dict['slices']
	dyn_range = fig_dict['dyn_range']
	val_range = fig_dict['val_range']
	my_color_map = fig_dict['color']
	plotter = twplot.plotter(file_to_plot,units=fig_dict['units'])
	c = slicing_spec.find('t')
	if (c==2 or c==3) and slices[c-2]==':':
		gen_movie(primitive_file+'.gif',plotter,slicing_spec,slices,dyn_range,val_range,my_color_map)
		return primitive_file+'.gif'
	else:
		gen_plot(primitive_file+'.png',plotter,slicing_spec,tuple(map(int,slices)),dyn_range,val_range,my_color_map)
		return primitive_file+'.png'

def check_data(ci_dict):
	'''Read the data file and compare to expectations'''
	file_to_check = ci_dict['data']
	val_range = ci_dict['range']
	tolerance = ci_dict['tol']
	dat = np.load(file_to_check)
	if np.abs(np.max(dat[-1,...]) - val_range[1]) > tolerance:
		return 1
	if np.abs(np.min(dat[-1,...]) - val_range[0]) > tolerance:
		return 1
	return 0

def comment_remover(text):
	'''Function to strip comments from TW input file
	From stackoverflow.com, author Markus Jarderot'''
	def replacer(match):
		s = match.group(0)
		if s.startswith('/'):
			return " " # note: a space and not an empty string
		else:
			return s
	pattern = re.compile(
		r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
		re.DOTALL | re.MULTILINE
	)
	return re.sub(pattern, replacer, text)

def parse_input_file(ex_path):
	'''Put important info from input file into a dictionary'''
	input_dict = {'ci' : [], 'fig': []}
	with open(ex_path) as f:
		for line in f:
			c = line.find('CITEST')
			if c>=0:
				args = line[c:].split()
				if len(args)!=4:
					raise ValueError('Example file with badly formed CITEST specification')
				input_dict['ci'] += [{}]
				input_dict['ci'][-1]['data'] = args[1]
				for s in args[2:]:
					if s.split('=')[0]=='range':
						input_dict['ci'][-1]['range'] = tuple(map(float,s.split('=')[1].split(',')))
					if s.split('=')[0]=='tolerance':
						input_dict['ci'][-1]['tol'] = float(s.split('=')[1])
				if len(input_dict['ci'][-1])!=3:
					raise ValueError('Example file with badly formed CITEST specification')
			c = line.find('TWTEST')
			if c>=0:
				args = line[c:].split()
				if args[1]!='matplotlib':
					raise ValueError('Example file with invalid plotter'+args[1])
				input_dict['fig'] += [{}]
				input_dict['fig'][-1]['slicing_spec'] = args[2].split('=')[0]
				slices = []
				for s in args[2].split('=')[1].split(','):
					slices.append(s)
				input_dict['fig'][-1]['slices'] = slices
				input_dict['fig'][-1]['data'] = args[3]
				input_dict['fig'][-1]['dyn_range'] = 0.0
				input_dict['fig'][-1]['val_range'] = (0.0,0.0)
				input_dict['fig'][-1]['color'] = 'viridis'
				input_dict['fig'][-1]['units'] = 'plasma'
				if len(args)>4:
					for keyval in args[4:]:
						key = keyval.split('=')[0]
						val = keyval.split('=')[1]
						if key=='dr':
							input_dict['fig'][-1]['dyn_range'] = np.float(val)
						if key=='range':
							r = val.split(',')
							input_dict['fig'][-1]['val_range'] = ( np.float(r[0]) , np.float(r[1]) )
						if key=='color':
							input_dict['fig'][-1]['color'] = val
						if key=='units':
							input_dict['fig'][-1]['units'] = val
	if len(input_dict['fig'])==0 and len(input_dict['ci'])==0:
		return input_dict
	with open(ex_path) as f:
		data = f.read()
		data = comment_remover(data)
		words = data.replace("="," ").replace(","," ").replace("("," ").replace(")"," ").split()
		start_looking = False
		user_vars = {}
		for i,word in enumerate(words):
			if word=="#define":
				user_vars[words[i+1]] = words[i+2]
			if word=='new' and words[i+1]=='grid':
				start_looking = True
			if start_looking and word=='dimensions':
				subs = []
				for d in range(1,4):
					if words[i+d] in user_vars:
						subs += [user_vars[words[i+d]]]
					else:
						subs += [words[i+d]]
				input_dict['dims'] = (int(subs[0]),int(subs[1]),int(subs[2]))
				start_looking = False
	return input_dict

def get_requested_concurrency(arg_list):
	mpi_procs = 1
	omp_threads = 1
	for i,arg in enumerate(arg_list):
		if arg=='-n' or arg=='-np':
			mpi_procs = int(arg_list[i+1])
		if arg=='-c':
			omp_threads = int(arg_list[i+1])
	return mpi_procs,omp_threads

def optimize_concurrency(num_procs,num_threads,dims):
	num_dims = 0
	dim1 = 0
	if dims[0]>1:
		dim1 = dims[0]
		num_dims += 1
	if dims[1]>1:
		dim1 = dims[1]
		num_dims += 1
	if dims[2]>1:
		dim1 = dims[2]
		num_dims += 1
	if num_dims==1:
		req_nodes = num_procs*num_threads
		mpi_nodes = req_nodes
		while dim1%(2*mpi_nodes)!=0:
			mpi_nodes -= 1
		return mpi_nodes,1
	else:
		return num_procs,num_threads

def form_command(arg_list,num_procs,num_threads):
	cmd_list = []
	for arg in arg_list:
		cmd_list += [arg]
	for i,arg in enumerate(cmd_list):
		if arg=='-n' or arg=='-np':
			cmd_list[i+1] = str(num_procs)
		if arg=='-c':
			cmd_list[i+1] = str(num_threads)
	cmd = ''
	for s in cmd_list:
		cmd += s + ' '
	return cmd[:-1]

def SeaTrials(args):
	'''Runs examples with TWTEST line and generates human-readable report.
	Arguments:
	args = dictionary of preprocessed command line arguments'''

	mpl.rcParams.update({'text.usetex' : False , 'font.size' : 16})

	html_doc = ''
	html_doc += '<!DOCTYPE html>'
	html_doc += '<html><head><title>TurboWAVE Report</title></head>'
	html_doc += '<body>'

	html_doc += '<h1 style="background-color:black;color:white;">TurboWAVE Test Suite Report</h1>\n\n'

	html_doc += '<h2 style="background-color:rgb(0,20,100);color:white;">Version Information</h2>\n\n'
	html_doc += '<p>Date of report : ' + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	mpi_procs,omp_threads = get_requested_concurrency(args['--command'])
	html_doc += '<p>Requested MPI processes = ' + str(mpi_procs) + '</p>'
	html_doc += '<p>Requested OMP threads = ' + str(omp_threads) + '</p>'

	cleanup('*.gif')
	cleanup('*.png')
	cleanup('*.npy')

	save_cwd = os.getcwd()
	os.chdir(args['--root'])
	try:
		print('Running Sea Trials')
		print('This may take a long time...')
		print()
		compl = subprocess.run(["git","status"],stdout=subprocess.PIPE,universal_newlines=True)
		html_doc += '<p>Git status:</p>'
		html_doc += '<blockquote><pre>'+compl.stdout+'</pre></blockquote>'

		compl = subprocess.run(["git","log"],stdout=subprocess.PIPE,universal_newlines=True)
		html_doc += '<p>Git last commit:</p>'
		gitloglines = compl.stdout.splitlines(keepends=True)
		html_doc += '<blockquote><pre>'
		commit_count = 0
		for l in gitloglines:
			if l[:6]=='commit':
				commit_count += 1
			if commit_count==2:
					break
			html_doc += l
		html_doc += '</pre></blockquote>'
	except FileNotFoundError:
		html_doc += '<p>ERROR: Could not find git (used to report versioning information).</p>'
	os.chdir(save_cwd)

	try:
		category_path_list = glob.glob(args['--root']+'/core/examples/*')
		category_path_list = sorted(category_path_list,key=lambda s: ntpath.basename(s)[0])
		category_list = []
		for s in category_path_list:
			category_list.append(ntpath.basename(s))
		if len(args['--categories'])>0:
			for req in args['--categories']:
				if req not in category_list:
					raise ValueError('Requested category '+req+' not found.')
			category_list = args['--categories']

		for cat in category_list:
			print('=====================================')
			print('Category',cat)
			html_doc += '\n\n<h2 style="background-color:rgb(0,20,100);color:white;">"' + cat + '" Subdirectory</h2>\n\n'
			ex_path_list = glob.glob(args['--root']+'/core/examples/'+cat+'/*')
			ex_path_list = sorted(ex_path_list,key=lambda s: ntpath.basename(s)[0])
			ex_list = []
			for s in ex_path_list:
				ex_list.append(ntpath.basename(s))
			for i,ex_path in enumerate(ex_path_list):
				print('--------------------------------------')
				print('Example',ex_list[i])
				html_doc += '\n<h3 style="background-color:rgb(100,250,200);">"' + ex_list[i] + '" Example</h3>\n'

				input_dict = parse_input_file(ex_path)

				if len(input_dict['fig'])>0:
					# Run this case
					nprocs,nthreads = optimize_concurrency(mpi_procs,omp_threads,input_dict['dims'])
					cmd = form_command(args['--command'],nprocs,nthreads)
					cmd += ' --no-interactive --input-file ' + ex_path
					print('Executing',cmd,'...')
					html_doc += '<p>TW command line = <samp>'+cmd+'</samp></p>'
					compl = subprocess.run(cmd.split(),stdout=subprocess.PIPE,universal_newlines=True)

					# Record the results
					if compl.returncode==0:
						warnings = []
						for l in compl.stdout.splitlines():
							if l.find('WARNING')>=0:
								warnings.append(l)
						if len(warnings)>0:
							print('Completed successfully but with warnings:')
							html_doc += '<p>Completed successfully but with warnings:</p>'
							for w in warnings:
								print(w)
								html_doc += '<p><samp>' + w.replace('<','&lt;').replace('>','&gt;') + '</samp></p>'
						else:
							print('Completed successfully.')
							html_doc += '<p>Completed successfully.</p>'
						html_doc += '<p>Representative figure:</p>'
						print('Generating visualization...',end='',flush=True)
						primitive_file = cat + '-' + ex_list[i] + '-' + input_dict['fig'][-1]['data']
						try:
							viz_file = gen_viz(primitive_file,input_dict['fig'][-1])
							html_doc += '<p><img src="' + viz_file + '" alt="Visualization is missing."</p>'
							print('OK')
						except OSError as err:
							print("Trouble")
							print(err)
							html_doc += '<p>There was a problem creating the visualization.</p>'
					else:
						print('Did not complete successfully.  Error report from TW follows:')
						html_doc += '<p>Did not complete successfully.  Error report from TW follows:</p>'
						for l in compl.stdout.splitlines():
							if l.find('ERROR')>=0:
								print(l)
								html_doc += '<p><samp>' + l.replace('<','&lt;').replace('>','&gt;') + '</samp></p>'
					cleanup('*.npy')
					cleanup('*tw_metadata*')
					cleanup('*grid_warp*')
					if len(glob.glob('twstat'))==1:
						os.remove('twstat')
				else:
					print('  Test not requested, or not a TW input file.')
					html_doc += '<p>Test not requested, or not a TW input file.</p>'


	except:
		traceback.print_exc()
		print('Unrecoverable error, attempting to close report...')
		html_doc += '\n\n<h2>REPORT ENDS PREMATURELY</h2>\n\n'

	html_doc += '</body></html>'

	with open('twreport.html','w') as f:
		print(html_doc,file=f)

def UnitTest(args):
	'''Runs turboWAVE unit tests.
	Arguments:
	args = dictionary of preprocessed command line arguments'''

	err_report = {}
	mpi_procs,omp_threads = get_requested_concurrency(args['--command'])

	try:
		print('Running all Unit Tests')
		print('----------------------')
		cmd = ['tw3d','-n','2','--unit-test','--all']
		compl = subprocess.run(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE,universal_newlines=True)

		# Check the results
		if compl.returncode==0:
			passing = compl.stdout.count(term.ok)
			missing = compl.stdout.count(term.yellow)
			failing = compl.stdout.count(term.err)
			print('    ' + term.ok + ' ' + term.green + 'passing' + term.reset_all)
			print('    {} passing, {} failing, {} missing'.format(passing,failing,missing))
		else:
			print('    ' + term.err + ' ' + term.red + 'failure' + term.reset_all)
			err_report['unit tests'] = {}
			err_report['unit tests']['stdout'] = compl.stdout
			err_report['unit tests']['stderr'] = compl.stderr

		# Error report
		for i,err in enumerate(err_report):
			print()
			print(str(i+1)+'. '+term.red + err + term.reset_all + ' is failing.')
			print('    stdout follows:')
			for line in err_report[err]['stdout'].splitlines():
				print('    ' + term.cyan + line + term.reset_all)
			print('    stderr follows:')
			for line in err_report[err]['stderr'].splitlines():
				print('    ' + term.yellow + line + term.reset_all)

	except:
		traceback.print_exc()
		print('unrecoverable error')

def CITest(args):
	'''Runs tests with CITEST line and spot checks results.
	Arguments:
	args = dictionary of preprocessed command line arguments'''

	err_report = {}
	mpi_procs,omp_threads = get_requested_concurrency(args['--command'])

	cleanup('*.gif')
	cleanup('*.png')
	cleanup('*.npy')

	try:
		print('Running Integration Tests')
		print('-------------------------')
		category_path_list = glob.glob(args['--root']+'/core/test/*')
		category_path_list = sorted(category_path_list,key=lambda s: ntpath.basename(s)[0])
		category_list = []
		for s in category_path_list:
			category_list.append(ntpath.basename(s))
		if len(args['--categories'])>0:
			for req in args['--categories']:
				if req not in category_list:
					raise ValueError('Requested category '+req+' not found.')
			category_list = args['--categories']

		for cat in category_list:
			print('Category',cat)
			ex_path_list = glob.glob(args['--root']+'/core/test/'+cat+'/*')
			ex_path_list = sorted(ex_path_list,key=lambda s: ntpath.basename(s)[0])
			ex_list = []
			for s in ex_path_list:
				ex_list.append(ntpath.basename(s))
			for i,ex_path in enumerate(ex_path_list):

				input_dict = parse_input_file(ex_path)

				if len(input_dict['ci'])>0:
					# Run this case
					nprocs,nthreads = optimize_concurrency(mpi_procs,omp_threads,input_dict['dims'])
					cmd = form_command(args['--command'],nprocs,nthreads)
					cmd += ' --no-interactive --input-file ' + ex_path
					compl = subprocess.run(cmd.split(),stdout=subprocess.PIPE,stderr=subprocess.PIPE,universal_newlines=True)

					# Check the results
					if compl.returncode==0:
						try:
							if check_data(input_dict['ci'][-1])==1:
								print('    ' + term.err + ' ' + term.red + ex_list[i] + term.reset_all)
								err_report[ex_list[i]] = {}
								err_report[ex_list[i]]['stdout'] = ''
								err_report[ex_list[i]]['stderr'] = input_dict['ci'][-1]['data'] + ' extremum delta exceeds tolerance'
							else:
								print('    ' + term.ok+ ' ' + term.green + ex_list[i] + term.reset_all)
						except OSError as err:
							err_report[ex_list[i]] = {}
							err_report[ex_list[i]]['stdout'] = ''
							err_report[ex_list[i]]['stderr'] = err
					else:
						print('    ' + term.err + ' ' + term.red + ex_list[i] + term.reset_all + ' (run failed)')
						err_report[ex_list[i]] = {}
						err_report[ex_list[i]]['stdout'] = compl.stdout
						err_report[ex_list[i]]['stderr'] = compl.stderr

					cleanup('*.npy')
					cleanup('*tw_metadata*')
					cleanup('*grid_warp*')
					if len(glob.glob('twstat'))==1:
						os.remove('twstat')

		# Error report
		for i,err in enumerate(err_report):
			print()
			print(str(i+1)+'. '+term.red + err + term.reset_all + ' is failing.')
			print('    stdout follows:')
			for line in err_report[err]['stdout'].splitlines():
				print('    ' + term.cyan + line + term.reset_all)
			print('    stderr follows:')
			for line in err_report[err]['stderr'].splitlines():
				print('    ' + term.yellow + line + term.reset_all)

	except:
		traceback.print_exc()
		print('unrecoverable error')

def panic(mess):
	print(term.red+'ERROR: '+term.reset_all+mess)
	print('suggestion '+term.bold+term.green+'twtest --help'+term.reset_all)
	exit(1)

def main():

	args = { '-v': None,
		'--version' : None,
		'-h' : None,
		'--help' : None,
		'--unit': False,
		'--integration': False,
		'--sea-trials': False,
		'--categories': [],
		'--root': '',
		'--command': [],
		'expecting_value': None
	}

	for arg in sys.argv[1:]:
		if arg not in args and args['expecting_value']==None:
			panic('argument '+arg+' not recognized')
		if args['expecting_value']!=None:
			if args['expecting_value']=='--categories':
				args['--categories'] = arg.split(',')
				args['expecting_value'] = None
			if args['expecting_value']=='--root':
				args['--root'] = arg
				args['expecting_value'] = None
			if args['expecting_value']=='--command':
				args['--command'] += [arg]
				# Consumes remaining arguments
		else:
			if arg=='--help' or arg=='-h':
				print_usage()
				exit(0)
			if arg=='--version' or arg=='-v':
				print_version()
				exit(0)
			if arg in ['--integration','--unit','--sea-trials']:
				args[arg] = True

			if arg in ['--categories','--root','--command']:
				args['expecting_value'] = arg

	if args['--root']=='':
		panic('root directory not specified')
	if len(args['--command'])==0:
		panic('command not specified')
	if not os.path.isdir(args['--root']):
		panic('requested root directory ('+args['--root']+') does not exist.')

	if args['--unit']:
		UnitTest(args)
	if args['--integration']:
		print()
		CITest(args)
	if args['--sea-trials']:
		print()
		SeaTrials(args)



from setuptools import setup
from setuptools import find_packages

with open('README.md','r',encoding='utf-8') as fh:
	long_description = fh.read()

setup(name='twutils',
	version='4.7.0b',
	license='PUBLIC-DOMAIN',
	author='Daniel Gordon',
	author_email='daniel.gordon@nrl.navy.mil',
	description='TurboWAVE utilities',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/USNavalResearchLaboratory/turboWAVE',
	classifiers=[
		'Programming Language :: Python :: 3',
		],
	package_dir={'':'src'},
	packages=find_packages(where='src'),
	python_requires='>=3.8',
	install_requires=['scipy>=1.5','matplotlib>=3.3','jupyter>=1','ipympl>=0.5','pillow>=8','h5py'],
	include_package_data=True,
	entry_points = { 'console_scripts': [
		'twinstall=twutils.command_line.twinstall:main',
		'twplot=twutils.command_line.twplot:main',
		'twtest=twutils.command_line.twtest:main',
		'twscan=twutils.command_line.twscan:main',
		'os2tw=twutils.command_line.os2tw:main',
		'wigner=twutils.command_line.wigner:main'] },
)



import sys
import glob
import os
import pkg_resources
import PIL.Image
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import twutils.plot as twplot
from scipy import constants as C

# Plotter for publication quality figures or presentation quality movies

def print_usage():
	print('Usage: twplot slicing=slices/... file1,... [panels=a,b,...] [mult=1.0,1.0,1.0/...]')
	print('    [range=low,high/...] [layout=2x2] [dr=0.0,...] [color=viridis,...] [roi=h0,h1,v0,v1/...]')
	print('    [units=mks,...] [[tex]labels=hlab,vlab,clab/...]')
	print('------------------Examples----------------')
	print('2D zx-plot: python twplot.py zxyt=0,0 rho.npy')
	print('1D x-plot: python twplot.py xyzt=0,0,0 Ex.npy')
	print('Hybrid: python twplot.py zxyt=0,0/zxyt=0,0,0 rho.npy,Ex.npy panels=a,b layout=1x2')
	print('2D zx-movie over all time slices: python twplot.py zxyt=0,: Ex.npy')
	print('-------------------General Notes--------------------')
	print('Extra spaces (e.g. around commas or equals) are not allowed.')
	print('Displays any number of panels.')
	print('Required arguments are positional, optional arguments are key=value pairs.')
	print('------------------Required Arguments-----------------------')
	print('slicing: 4-character string, such as xyzt, where the first 1 or 2 axes are plotted.')
	print('slices: is a comma delimited list of slice indices.')
	print('   Number of slices determines whether the plot is 1D or 2D.')
	print('file1: name of the first file to plot, etc. File names cannot contain commas.')
	print('------------------Optional Arguments-----------------------')
	print('panels: labels for the subplot panels, e.g., a,b,etc.')
	print('mult: rescaling multipliers, multiplying haxis,vaxis,data.')
	print('range: explicit setting of the low and high bounds of plotting range.')
	print('layout: layout of the grid on which to put multiple plots.')
	print('dr: 0.0 signals full range on linear scale, any other float is the number of decades spanned.')
	print('color: any Matplotlib color map, search internet for list (e.g., viridis, jet, inferno, etc.)')
	print('roi: select a subdomain to plot, otherwise the full domain is plotted.')
	print('units: select from plasma,atomic,natural,mks,cgs')
	print('[tex]labels: fully customizable labels for axes and colorbar.')
	print('   Including prefix tex causes each string to be automatically enclosed in $ tokens')
	print('   Some characters require special treatment if you want to use them in a label.')
	print('   Enter SPACE,COMMA,SEMICOLON,or SLASH to get the corresponding character.')
	print('   Depending on shell, various characters may need to be escaped (KNOW YOUR SHELL).')
	print('----------------------Animations----------------------')
	print('Put a python range as one of the slices to generate animated GIF.')
	print('For example, zxyt=0,0,2:5 would animate time slices 2,3,4.')

def print_version():
    print('twplot is provided by twutils, version '+pkg_resources.get_distribution('twutils').version)

# Label scheme

def format_label(l):
	return l.replace('SPACE',' ').replace('COMMA',',').replace('SEMICOLON',';').replace('SLASH','/')

# Functions helpful for animations

def cleanup(wildcarded_path):
	cleanstr = glob.glob(wildcarded_path)
	for f in cleanstr:
		os.remove(f)

def ParseSlices(dims,ax_list,slice_str_list):
	'''Function to generate a list of slice tuples for the movie.
	dims = dimensions of all 4 axes
	ax_list = slicing_spec as list of integer axis identifiers.
	slice_str_list = list of slice strings, can be indices or ranges.
	Returns slice_tuples,movie.'''
	slice_tuples = []
	range_tuples = []
	movie = False
	# Construct list of range tuples
	sax = ax_list[4-len(slice_str_list):]
	for saxi,slice_str in enumerate(slice_str_list):
		rng = slice_str.split(':')
		tup = ()
		for i,el in enumerate(rng):
			if el=='' and i==0:
				el = '0'
			if el=='' and i==1:
				el = str(dims[sax[saxi]])
			if el=='' and i==2:
				el = '1'
			tup += (int(el),)
		range_tuples.append(tup)
	# Determine the range of the movie frames
	frame_rng = range(1)
	for rng in range_tuples:
		movie = movie or len(rng)>1
		if len(rng)==2:
			frame_rng = range(rng[0],rng[1])
		if len(rng)==3:
			frame_rng = range(rng[0],rng[1],rng[2])
	# Construct list of slice tuples
	for r in frame_rng:
		tup = ()
		for rng in range_tuples:
			if len(rng)>1:
				tup += (r,)
			else:
				tup += rng
		slice_tuples.append(tup)
	return slice_tuples,movie

def main():

	if len(sys.argv)<2 or '--help' in sys.argv or '-h' in sys.argv:
		print_usage()
		exit(0)

	if '--version' in sys.argv or '-v' in sys.argv:
		print_version()
		exit(0)

	# Matplotlib setup

	mpl.rcParams.update({'text.usetex' : False , 'font.size' : 10})
	proportional = False
	if proportional:
		my_aspect = 'equal'
	else:
		my_aspect = 'auto'

	# Process command line arguments and setup plotter object

	slicing_spec = []
	primitive_slices = []
	for spec in sys.argv[1].split('/'):
		slicing_spec.append(spec.split('=')[0])
		primitive_slices.append(spec.split('=')[1].split(','))
	file_to_plot = sys.argv[2].split(',')
	N = len(file_to_plot)
	panels = []
	rows = 1
	cols = N
	dyn_range = []
	color = []
	roi = []
	val_range = []
	mult = []
	units = []
	labels=[]
	texlabels=[]
	ask = 'yes'
	keylist = ['panels','layout','dr','color','roi','range','mult','units','labels','texlabels']
	for keyval in sys.argv[3:]:
		key = keyval.split('=')[0]
		arg = keyval.split('=')[1]
		if key not in keylist:
			raise ValueError('Invalid key in optional arguments.')
		if key=='panels':
			panels = arg.split(',')
		if key=='layout':
			layout = arg.split('x')
			rows = int(layout[0])
			cols = int(layout[1])
		if key=='dr':
			dyn_range = arg.split(',')
		if key=='color':
			color = arg.split(',')
		if key=='roi':
			roi = arg.split('/')
		if key=='range':
			val_range = arg.split('/')
		if key=='mult':
			mult = arg.split('/')
		if key=='units':
			units = arg.split(',')
		if key=='labels':
			labels = arg.split('/')
		if key=='texlabels':
			texlabels = arg.split('/')

	for i in range(N-len(slicing_spec)):
		slicing_spec.append(slicing_spec[-1])
		primitive_slices.append(primitive_slices[-1])
	for i in range(N-len(panels)):
		panels.append('')
	for i in range(N-len(dyn_range)):
		dyn_range.append('0')
	for i in range(N-len(val_range)):
		val_range.append('')
	for i in range(N-len(color)):
		color.append('viridis')
	for i in range(N-len(roi)):
		roi.append('')
	for i in range(N-len(mult)):
		mult.append('1.0,1.0,1.0')
	for i in range(N-len(labels)):
		labels.append(',,')
	for i in range(N-len(texlabels)):
		texlabels.append(',,')
	for i in range(N-len(units)):
		units.append('plasma')

	plotter = []
	for i,f in enumerate(file_to_plot):
		needs_buffer = slicing_spec[i][0]=='t' or (slicing_spec[i][1]=='t' and len(primitive_slices[i])==2)
		plotter.append(twplot.plotter(f,buffered=needs_buffer,units=units[i]))
		plotter[-1].display_info()

	# Set up animation slices

	slice_tuples = []
	movie = []
	for i in range(N):
		axes = twplot.get_axis_info(slicing_spec[i])
		dims = plotter[i].dims4()
		st,m = ParseSlices(dims,axes,primitive_slices[i])
		slice_tuples.append(st)
		movie.append(m)

	# Check existing files and clean if there is a movie

	if movie[0]:
		img_files = glob.glob('frame*.png')
		if len(img_files)>0 and ask=='yes':
			ans = ''
			while ans!='y' and ans!='n':
				ans = input('Found some frame*.png files, OK to clean (y/n) ?')
			if ans=='n':
				print('STOPPED. Please run script in a directory where there are no important files of the form frame*.png.')
				exit(1)

		for img_file in img_files:
			os.remove(img_file)

		mov_files = glob.glob('mov.gif')
		if len(mov_files)>0 and ask=='yes':
			ans = ''
			while ans!='y' and ans!='n':
				ans = input('Found mov.gif, OK to overwrite (y/n) ?')
			if ans=='n':
				print('STOPPED. Please do something with existing mov.gif and try again.')
				exit(1)

	# Determine the global color scale bounds

	global_min = 1e50*np.ones(N)
	global_max = -1e50*np.ones(N)
	for i,p in enumerate(plotter):
		if val_range[i]=='':
			for slice_now in slice_tuples[i]:
				dmult = float(mult[i].split(',')[2])
				if len(primitive_slices[i])==2:
					test_array,plot_dict = p.falsecolor2d(slicing_spec[i],slice_now,float(dyn_range[i]))
					local_min = plot_dict['vmin']*dmult
					local_max = plot_dict['vmax']*dmult
				if len(primitive_slices[i])==3:
					abcissa,test_array,plot_dict = p.lineout(slicing_spec[i],slice_now,float(dyn_range[i]))
					local_min = np.min(test_array)*dmult
					local_max = np.max(test_array)*dmult
				if local_min>local_max:
					local_min,local_max = local_max,local_min
				if local_min<global_min[i]:
					global_min[i] = local_min
				if local_max>global_max[i]:
					global_max[i] = local_max
		else:
			global_min[i] = float(val_range[i].split(',')[0])
			global_max[i] = float(val_range[i].split(',')[1])

	# Generate the movie or show a single frame.
	# The number of frames is dictated by the first panel.

	for file_idx in range(len(slice_tuples[0])):

		sz = (cols*5,rows*4)
		plt.figure(file_idx,figsize=sz,dpi=100)

		for i,p in enumerate(plotter):
			slice_now = slice_tuples[i][file_idx]
			exm = np.zeros(4)
			exm[:2] = float(mult[i].split(',')[0])
			exm[2:] = float(mult[i].split(',')[1])
			dmult = float(mult[i].split(',')[2])
			plt.subplot(rows,cols,i+1)
			if len(primitive_slices[i])==2:
				data_slice,plot_dict = p.falsecolor2d(slicing_spec[i],slice_now,float(dyn_range[i]))
				plt.imshow(data_slice*dmult,
					origin='lower',
					aspect=my_aspect,
					extent=np.array(plot_dict['extent'])*exm,
					vmin=global_min[i],
					vmax=global_max[i],
					cmap=color[i])
				b = plt.colorbar()
				l = labels[i].split(',')
				tl = texlabels[i].split(',')

				if l[0]=='' and tl[0]=='':
					plt.xlabel(plot_dict['xlabel'],size=12)
				if not l[0]=='':
					plt.xlabel(format_label(l[0]),size=12)
				if not tl[0]=='':
					plt.xlabel(format_label(r'$'+tl[0]+r'$'),size=12)

				if l[1]=='' and tl[1]=='':
					plt.ylabel(plot_dict['ylabel'],size=12)
				if not l[1]=='':
					plt.ylabel(format_label(l[1]),size=12)
				if not tl[1]=='':
					plt.ylabel(format_label(r'$'+tl[1]+r'$'),size=12)

				if l[2]=='' and tl[2]=='':
					b.set_label(plot_dict['blabel'],size=12)
				if not l[2]=='':
					b.set_label(format_label(l[2]),size=12)
				if not tl[2]=='':
					b.set_label(format_label(r'$'+tl[2]+r'$'),size=12)

				if roi[i]=='':
					roi_i = np.array(plot_dict['extent'])*exm
				else:
					roi_i = []
					for s in roi[i].split(','):
						roi_i.append(float(s))
					plt.xlim(roi_i[0],roi_i[1])
					plt.ylim(roi_i[2],roi_i[3])
				if not panels[i]=='':
					plt.text(roi_i[0],roi_i[3]+0.03*(roi_i[3]-roi_i[2]),'('+panels[i]+')')
			if len(primitive_slices[i])==3:
				abcissa,ordinate,plot_dict = p.lineout(slicing_spec[i],slice_now,float(dyn_range[i]))
				plt.plot(abcissa*exm[0],ordinate*exm[2])
				l = labels[i].split(',')
				tl = texlabels[i].split(',')

				if l[0]=='' and tl[0]=='':
					plt.xlabel(plot_dict['xlabel'],size=12)
				if not l[0]=='':
					plt.xlabel(format_label(l[0]),size=12)
				if not tl[0]=='':
					plt.xlabel(format_label(r'$'+tl[0]+r'$'),size=12)

				if l[1]=='' and tl[1]=='':
					plt.ylabel(plot_dict['ylabel'],size=12)
				if not l[1]=='':
					plt.ylabel(format_label(l[1]),size=12)
				if not tl[1]=='':
					plt.ylabel(format_label(r'$'+tl[1]+r'$'),size=12)

				if roi[i]=='':
					roi_i = [abcissa[0]*exm[0],abcissa[-1]*exm[0],global_min[i]*exm[2],global_max[i]*exm[2]]
				else:
					s = roi[i].split(',')
					roi_i = [float(s[0]),float(s[1]),float(s[2]),float(s[3])]
					plt.xlim(roi_i[0],roi_i[1])
					plt.ylim(roi_i[2],roi_i[3])
				if not panels[i]=='':
					plt.text(roi_i[0],roi_i[3]+0.03*(roi_i[3]-roi_i[2]),'('+panels[i]+')')

			plt.tight_layout()

		if movie[0]:
			img_file = 'frame{:03d}.png'.format(file_idx)
			print('saving',img_file,'...')
			plt.savefig(img_file)
			plt.close()

	if movie[0]:
		print('Consolidating into movie file...')
		images = []
		frameRateHz = 5
		img_files = sorted(glob.glob('frame*.png'))
		for f in img_files:
			images.append(PIL.Image.open(f))
		images[0].save('mov.gif',save_all=True,append_images=images[1:],duration=int(1000/frameRateHz),loop=0)
		cleanup('frame*.png')
		print('Done.')
	else:
		plt.show()



'''
Module :samp:`twutils.QO.ionization`
====================================

This module can be used to compute ionization rates that may be needed in pre or post processing.  It contains dictionaries of atomic data, and classes for computing ionization rates based on various models appearing in the literature.  The ionization classes use atomic units, but include routines for converting to and from mks units.

References
----------

The ionization models herein are detailed in the following references.

#. L.V. Keldysh, Sov. Phys. JETP 20 (5), 1307 (1965)

#. A.M. Perelemov, V.S. Popov, M.V. Terent'ev, Sov. Phys. JETP 23 (5), 924 (1966)

#. M.V. Ammosov, N.B. Delone, V.P. Krainov, Sov. Phys. JETP 64 (6), 1191 (1986)

#. G.L. Yudin and M.Y. Ivanov, Phys. Rev. A 64, 013409 (2001)

#. S.V. Popruzhenko, V.D. Mur, V.S. Popov, D. Bauer, Phys. Rev. Lett. 101, 193003 (2008)

#. M\. Klaiber, E. Yakaboylu, K.Z. Hatsagortsyan, Phys. Rev. A 87, 023417 and 023418 (2013)

Constants
----------

* Dictionary ``ip``: keys are atomic symbols (e.g. 'H','He',etc.), values are lists of ionization potentials in eV.

* Dictionary ``atomic_mass``: keys are atomic symbols, values are atomic masses in proton mass units.

* Dictionary ``atomic_number`` : keys are atomic symbols, values are the number of protons.

An example of usage is as follows::

	import twutils.QO.ionization as iz

	print('argon ionization potential =',iz.ip['Ar'][0])
	print('argon[3+] ionization potential =',iz.ip['Ar'][3])
	print('atomic mass of argon is',iz.atomic_mass['Ar'])

Classes and functions
----------------------
'''

import numpy as np
from scipy import special as spsf
from scipy import constants as C
import scipy.integrate


# H,He,Li,Be,B,C,N,O,Ne,Ar data from photocopy of three pages (10-205--10-207) of unknown book
# Krypton I-XXVI, J. Sugar, A. Musgrove, J. Phys. Chem. Ref. Data 20, 859 (1991)
# Krypton V-XXVI, T. Shirai, K. Okazaki, and J. Sugar, J. Phys. Chem. Ref. Data 24, 1577 (1995)
# Xenon, E.B. Saloman, J. Phys. Chem. Ref. Data 33, 765 (2004)

ip = {'H' : [13.59844],
	'He' : [24.58741,54.41778],
	'Li' : [5.39172,75.64018,122.45429],
	'Be' : [9.32263,18.21116,153.89661,217.71865],
	'B' : [8.29803,25.15484,37.93064,259.37521,340.22580],
	'C' : [11.26030,24.38332,47.8878,64.4939,392.087,489.9934],
	'N' : [14.53414,29.6013,47.44924,77.4735,97.8902,552.0718,667.046],
	'O' : [13.61806,35.11730,54.9355,77.41353,113.8990,138.1197,739.29,871.4101],
	'Ne' : [21.56454,40.96328,63.45,97.12,126.21,157.93,207.2759,239.0989,1195.8286,1362.1995],
	'Ar' : [15.75962,27.62967,40.74,59.81,75.02,91.009,124.323,143.460,422.45,478.69,
		538.96,618.26,686.10,755.74,854.77,918.03,4120.8857,4426.2296],
	'Kr' : [13.99961,24.35985,36.950,52.5,64.7,78.5,111.0,125.802,230.85,268.2,
		308.2,350.1,390.9,446.6,491.8,540.7,591.5,641,785.9,833.0,
		883.9,936.7,997.7,1050.9,1151.4,1205.3,2928,3070,3227,3381,
		3594,3760,3966,4108.54,17297.16,17936.21],
	'Xe' : [12.130,20.975,31.05,40.9,54.14,66.703,91.6,105.976,179.85,201.7,
		229.02,263.5,294.4,325.3,358.3,389.6,420.9,452.2,572.5,607.7,
		642.9,678.1,726.0,762.4,852.7,857.0,1490,1491,1587,1684,
		1781,1877,1987,2085,2183,2281,2548,2637,2726,2814,
		3001,3093,3296,3334,7224,7491,7758,8024,8617,8899,
		9607,9813,40273,41300]}

atomic_mass = { 'H' : 1.00794 , 'He' : 4.002602 , 'Li' : 6.941 ,
	'Be' : 9.012182 , 'B' : 10.811 , 'C' : 12.0107 , 'N' : 14.0067 ,
	'O' : 15.9994 , 'F' : 18.9984 , 'Ne' : 20.1797 , 'Na' : 22.98977 ,
	'Mg' : 24.3050 , 'Al' : 26.9815 , 'Si' : 28.0855 , 'P' : 30.9738 ,
	'S' : 32.065 , 'Cl' : 35.453 , 'K' : 39.0983 , 'Ar' : 39.948 ,
	'Ca' : 40.078 , 'Sc' : 44.956 , 'Ti' : 47.867 , 'V' : 50.9415 ,
	'Cr' : 51.9961 , 'Mn' : 54.938 , 'Fe' : 55.845 , 'Co' : 58.933 ,
	'Ni' : 58.6934 , 'Cu' : 63.546 , 'Zn' : 65.409 , 'Ga' : 69.723 ,
	'Ge' : 72.64 , 'As' : 74.92160 , 'Se' : 78.96 , 'Br' : 79.904 ,
	'Kr' : 83.798 , 'Rb' : 85.4678 , 'Sr' : 87.62 , 'Y' : 88.90585 ,
	'Zr' : 91.224 , 'Nb' : 92.906 , 'Mo' : 95.94 , 'Tc' : 98 ,
	'Ru' : 101.07 , 'Rh' : 102.905 , 'Pd' : 106.42 , 'Ag' : 107.8682 ,
	'Cd' : 112.411 , 'In' : 114.818 , 'Sn' : 118.710 , 'Sb' : 121.760 ,
	'Te' : 127.60 , 'I' : 126.904 , 'Xe' : 131.293 }

atomic_number = { 'H' : 1 , 'He' : 2 , 'Li' : 3 ,
	'Be' : 4 , 'B' : 5 , 'C' : 6 , 'N' : 7 ,
	'O' : 8 , 'F' : 9 , 'Ne' : 10 , 'Na' : 11 ,
	'Mg' : 12 , 'Al' : 13 , 'Si' : 14 , 'P' : 15 ,
	'S' : 16 , 'Cl' : 17 , 'K' : 19 , 'Ar' : 18 , # n.b. ordered by mass
	'Ca' : 20 , 'Sc' : 21 , 'Ti' : 22 , 'V' : 23 ,
	'Cr' : 24 , 'Mn' : 25 , 'Fe' : 26 , 'Co' : 27 ,
	'Ni' : 28 , 'Cu' : 29 , 'Zn' : 30 , 'Ga' : 31 ,
	'Ge' : 32 , 'As' : 33 , 'Se' : 34 , 'Br' : 35 ,
	'Kr' : 36 , 'Rb' : 37 , 'Sr' : 38 , 'Y' : 39 ,
	'Zr' : 40 , 'Nb' : 41 , 'Mo' : 42 , 'Tc' : 43 ,
	'Ru' : 44 , 'Rh' : 45 , 'Pd' : 46 , 'Ag' : 47 ,
	'Cd' : 48 , 'In' : 49 , 'Sn' : 50 , 'Sb' : 51 ,
	'Te' : 52 , 'I' : 53 , 'Xe' : 54 }

def Shifted_Uion_H(E):
	"""Returns the Stark shifted ionization energy of Hydrogen"""
	return 0.5 + (9.0/4.0)*E**2 + 55.55*E**4 + 4908.0*E**6 + 794200*E**8

def Keldysh_Parameter(omega,Uion,E):
	'''Compute the adiabaticity parameter (Keldysh parameter)

	:param double omega: radiation frequency in atomic units
	:param double Uion: ionization potential in atomic units
	:param double E: electric field in atomic units

	:return: dimensionless adiabaticity parameter'''
	return omega*np.sqrt(2.0*Uion)/E

def CycleAveragingFactor(Uion,E):
	'''Returns the PPT cycle averaging factor, multiply static rate by this factor to get average rate.

	:param double Uion: ionization potential in atomic units
	:param double E: electric field in atomic units'''
	return np.sqrt(3.0/np.pi) * np.sqrt(E) / (2*Uion)**0.75

class AtomicUnits:
	'''Class for converting to and from atomic units. Ionization classes derive from this.'''
	def __init__(self):
		'''Create an AtomicUnits object'''
		# atomic units in mks units
		self.xa = C.hbar / (C.m_e*C.c*C.alpha)
		self.ta = C.hbar / (C.m_e*C.c**2*C.alpha**2)
		self.ua = C.m_e*C.c**2*C.alpha**2
		self.ea = self.ua/self.xa/C.e
	def energy_to_eV(self,u):
		'''Convert an energy from atomic units to eV'''
		return u*self.ua/C.e
	def length_to_mks(self,x):
		'''Convert a length from atomic to mks units'''
		return x*self.xa
	def rate_to_mks(self,rate):
		'''Convert a time from atomic to mks units'''
		return rate/self.ta
	def time_to_mks(self,t):
		'''Convert a time from atomic to mks units'''
		return t*self.ta
	def field_to_mks(self,field):
		'''Convert electric field from atomic to mks units'''
		return field*self.ea
	def energy_to_au(self,u):
		'''Convert an energy in eV to atomic units'''
		return u*C.e/self.ua
	def length_to_au(self,x):
		'''Convert a length from mks to atomic units'''
		return x/self.xa
	def rate_to_au(self,rate):
		'''Convert a time from mks to atomic units'''
		return rate*self.ta
	def time_to_au(self,t):
		'''Convert a time from mks to atomic units'''
		return t/self.ta
	def field_to_au(self,field):
		'''Convert electric field from mks to atomic units'''
		return field/self.ea

class Ionization(AtomicUnits):
	'''Base class for computing ionization rates'''
	def __init__(self,averaging,Uion,Z,lstar=0,l=0,m=0,w0=0,terms=1):
		'''Create an ionization object.  Base class should be treated as abstract.

		:param bool averaging: whether this object will cycle average the rate, or try to return static rate
		:param double Uion: ionization potential in atomic units
		:param double Z: residual charge in atomic units
		:parameter float lstar: effective orbital angular momentum
		:parameter int l: orbital angular momentum quantum number
		:parameter int m: orbital angular momentum projection in direction of field
		:parameter float w0: angular frequency of radiation in atomic units
		:parameter int terms: terms to keep in expansion over multiphoton orders'''
		super().__init__()
		self.avg = averaging
		self.Uion = Uion
		self.Z = Z
		self.lstar = lstar
		self.l = l
		self.m = m
		self.w0 = w0
		self.terms = terms
		self.cutoff_field = 1e-3
		# coefficients for standard form ionization formulas
		self.C_pre = 0.0
		self.C_pow = 0.0
		self.C_exp = 0.0
	def ToAverage(self):
		'''Switch to averaging.  This must not be called more than once.'''
		self.C_pre *= np.sqrt(3.0/np.pi) / (2*self.Uion)**0.75
		self.C_pow += 0.5
	def ToStatic(self):
		'''Switch to static rate.  This must not be called more than once.'''
		self.C_pre *= (2*self.Uion)**0.75 / np.sqrt(3.0/np.pi)
		self.C_pow -= 0.5
	def ExtractEikonalForm(self,E,dt,w00=0.0,bandwidth=1.0):
		"""Extract amplitude, phase, and center frequency from a carrier resolved field E.
		The assumed form is :math:`E = \Re\{Ae^{i\phi}\}`.

		:param numpy.array E: Electric field, any shape, axis 0 is time.
		:param double dt: time step.
		:param double w00: Center frequency, if zero deduce from the data (default).
		:param double bandwidth: relative bandwidth to keep, if unity no filtering (default)."""
		Nt = E.shape[0]
		ndims = len(E.shape)
		dw = 2*np.pi/(Nt*dt)
		# Get the complexified spectrum and carrier frequency
		Ew = np.fft.fft(E,axis=0)
		Ew[np.int(Nt/2):,...] = 0.0 # eliminate negative frequencies before extracting carrier
		if w00==0.0:
			# Compute a suitable carrier frequency based on intensity weighting
			if ndims==1:
				indices = np.arange(0,Nt)
			else:
				indices = np.einsum('i,jk',np.arange(0,Nt),np.ones(Ew.shape[1:]))
			carrier_idx = np.int(np.average(indices,weights=np.abs(Ew)**2))
			w0 = carrier_idx*dw
		else:
			# Carrier frequency is set by caller
			w0 = w00
			carrier_idx = np.int(w0/dw)
		# Bandpass filtering
		if bandwidth!=1.0:
			low_idx = np.int((w0-0.5*bandwidth*w0)/dw)
			high_idx = np.int((w0+0.5*bandwidth*w0)/dw)
			Ew[:low_idx] = 0.0
			Ew[high_idx:] = 0.0
		# Form the complex envelope in time
		Ew = np.roll(Ew,-carrier_idx,axis=0)
		Et = 2*np.fft.ifft(Ew,axis=0)
		# Get amplitude and phase
		amp = np.abs(Et)
		phase = (np.angle(Et).swapaxes(0,ndims-1) + w0*np.linspace(0,(Nt-1)*dt,Nt)).swapaxes(0,ndims-1)
		return amp,phase,w0
	def Rate(self,Es):
		'''Get the ionization rate, returned value is in atomic units.
		The object's internal state decides whether this is a static or cycle averaged rate.

		:param np.array Es: Time domain electric field in atomic units, any shape.  If a complex envelope is given :math:`|Es|` is expected to give the crest.'''
		E = np.abs(Es) + self.cutoff_field
		return self.C_pre * E**self.C_pow * np.exp(self.C_exp/E)

class Hydrogen(Ionization):
	"""Landau & Lifshitz formula for hydrogen, using PPT averaging"""
	def __init__(self,averaging,Uion,Z):
		super().__init__(averaging,Uion,Z)
		self.C_pre = 4.0
		self.C_pow = -1.0
		self.C_exp = -2.0/3.0
		if averaging:
			self.ToAverage()

class ADK(Ionization):
	'''Full ADK for arbitrary states'''
	def __init__(self,averaging,Uion,Z,lstar=0,l=0,m=0):
		super().__init__(averaging,Uion,Z,lstar,l,m)
		nstar = Z/np.sqrt(2*Uion)
		m = abs(self.m)
		e = np.exp(1.0)
		sn2l2 = np.sqrt(nstar**2-lstar**2)
		self.C_pre = np.sqrt(3/np.pi**3) * (2*l+1) * spsf.factorial(l+m) / spsf.factorial(m) / spsf.factorial(l-m)
		self.C_pre *= (e/sn2l2)**(m+1.5) # exponent is of poor print quality in JETP
		self.C_pre *= ((nstar+lstar)/(nstar-lstar))**(lstar+0.5)
		self.C_pre *= (Z**2/nstar**3)
		self.C_pre *= (4*e*Z**3/nstar**3/sn2l2)**(2*nstar-m-1.5)
		self.C_pow = m+1.5-2*nstar
		self.C_exp = -2*Z**3/(3*nstar**3) # ADK 1986 has nstar**4 in Eq. 21, must be a typo?
		if not averaging:
			self.ToStatic()

class ADK00(Ionization):
	'''ADK theory simplified for lstar=l=m=0'''
	def __init__(self,averaging,Uion,Z):
		super().__init__(averaging,Uion,Z)
		nstar = Z/np.sqrt(2*Uion)
		D = ((4.0*np.exp(1.0)*Z**3)/nstar**4)**nstar
		self.C_pre = D*D/(8.0*np.pi*Z)
		self.C_pow = 1.0-2*nstar # N.b. this is already for the static rate
		self.C_exp = -2*Z**3/(3.0*nstar**3)
		if averaging:
			self.ToAverage()

class KYH(Ionization):
	'''KYH theory for relativistic tunneling, using the dressed coulomb corrected SFA (Eq. 97)'''
	def __init__(self,averaging,Uion,Z):
		super().__init__(averaging,Uion,Z)
		Ua2 = Uion*C.alpha**2
		Ea = (2*Uion)**1.5 # definition of Ea applies to Eq. 97 but not Fig. 4 caption
		self.C_pre = 2**(3-4*Ua2)*np.sqrt(3/np.pi)*(1-7*Ua2/72)*np.exp(4*Ua2)*(2*Uion)**(1.75-3*Ua2)
		self.C_pre /= spsf.gamma(3-2*Ua2)
		self.C_pow = 2*Ua2 - 0.5
		self.C_exp = -(2.0/3.0)*Ea*(1-Ua2/12)
		if not averaging:
			self.ToStatic()

class PPT_Tunneling(Ionization):
	'''Evaluating the PPT theory in the tunneling limit, with the infinite sum solved analytically.
	This is PPT Eq. 59 multiplied by the standard Coulomb factor, and taking gamma=0.
	For the bound state coefficient we use ADK Eq. 19.'''
	def __init__(self,averaging,Uion,Z,lstar=0,l=0,m=0):
		super().__init__(averaging,Uion,Z,lstar,l,m)
		nstar = Z/np.sqrt(2*Uion)
		m = abs(self.m)
		F0 = Z**3/nstar**3
		# First without the Coulomb factor
		self.C_pre = Uion
		self.C_pre *= 2**(2*nstar) / (nstar*spsf.gamma(nstar+lstar+1)*spsf.gamma(nstar-lstar)) # |C|^2
		self.C_pre *= np.sqrt(6/np.pi)
		self.C_pre *= (2*l+1) * spsf.factorial(l+m) / 2**m / spsf.factorial(m) / spsf.factorial(l-m)
		self.C_pre *= (0.5/F0)**(m+1.5)
		self.C_pow = m+1.5
		self.C_exp = -2*F0/3
		# Account for Coulomb correction
		self.C_pre *= (2*F0)**(2*nstar)
		self.C_pow -= 2*nstar
		if not averaging:
			self.ToStatic()

class HybridTheories(Ionization):
	'''Base class for theories that handle both multiphoton and tunneling limits'''
	def __init__(self,averaging,Uion,Z,lstar=0,l=0,m=0,w0=0,terms=1):
		if m!=0.0:
			raise ValueError('Hybrid theories all require m=0')
		if w0==0.0:
			raise ValueError('Hybrid theories require w>0')
		super().__init__(averaging,Uion,Z,lstar,l,m,w0,terms)
	def Params(self,E,w,Uion,Z):
		nstar = Z/np.sqrt(2*Uion)
		F = E*nstar**3/Z**3
		gam = Z*w/(nstar*E)
		g = (3/(2*gam))*((1+1/(2*gam**2))*np.arcsinh(gam)-np.sqrt(1+gam**2)/(2*gam))
		nu = (self.Uion/w) * (1 + 1/(2*gam**2))
		return nstar,F,gam,g,nu
	def FourierSum(self,gam,nu):
		A0 = 0.0*gam; # make this the same shape as gam, if gam is an array
		alpha = 2 * (np.arcsinh(gam)-gam/np.sqrt(1+gam**2))
		beta = 2*gam/np.sqrt(1+gam**2)
		dnu = np.ceil(nu) - nu
		for n in range(self.terms):
			A0 += np.exp(-alpha*(n+dnu))*spsf.dawsn(np.sqrt(beta*(n+dnu))) # only valid for m=0
		return A0;

class PPT(HybridTheories):
	'''PPT theory accounting for multiphoton and tunneling limits, but assuming m=0.
	This is PPT Eq. 54 multiplied by the standard Coulomb factor, and taking m=0.
	For the bound state coefficient we use ADK Eq. 19.
	If static rate is requested, the result is only valid in the limit gamma=0'''
	def Rate(self,Es):
		E = np.abs(Es) + self.cutoff_field
		w = self.w0
		lstar = self.lstar
		l = self.l
		nstar,F,gam,g,nu = self.Params(E,w,self.Uion,self.Z)
		ans = self.Uion
		ans *= 2**(2*nstar) / (nstar*spsf.gamma(nstar+lstar+1)*spsf.gamma(nstar-lstar)) # |C|^2
		ans *= np.sqrt(6/np.pi) * (2*l+1)
		ans *= (F*np.sqrt(1+gam**2)/2)**1.5
		ans *= (4/np.sqrt(3*np.pi)) * (gam**2/(1+gam**2)) * self.FourierSum(gam,nu)
		ans *= np.exp(-2*g/(3*F))
		# Following is the Coulomb correction factor
		ans *= (2/F)**(2*nstar)
		if self.avg:
			return ans
		else:
			# If explicit, simply unwind cycle averaging.
			# This does not account for changes in sub-cycle structure with adiabaticity, but gives right average behavior.
			return ans/CycleAveragingFactor(self.Uion,E)

class PMPB(HybridTheories):
	'''PMBP theory, similar to PPT but with improved Coulomb factor, assuming m=0.
	Take lstar=0,l=0,m=0 to strictly recover PMPB Eq. 6.'''
	def Rate(self,Es):
		# Note that |C_PPT|^2 = 4*|C_PMPB|^2
		# This can be seen in how the respective authors write the wavefunction.
		E = np.abs(Es) + self.cutoff_field
		w = self.w0
		lstar = self.lstar
		l = self.l
		nstar,F,gam,g,nu = self.Params(E,w,self.Uion,self.Z)
		ans = (2/np.pi)*self.Uion
		ans *= 2**(2*nstar-2) / (nstar*spsf.gamma(nstar+lstar+1)*spsf.gamma(nstar-lstar)) # |C_PMBP|^2 generalized
		ans *= 2*l + 1 # generalize also for l>0, but demand m=0
		ans *= (self.Uion/w)**-1.5
		ans *= np.sqrt(2*gam/np.sqrt(1+gam**2))
		ans *= self.FourierSum(gam,nu)
		ans *= np.exp(-2*g/(3*F))
		# Following is the improved Coulomb correction factor
		ans *= (2/F)**(2*nstar) * (1+gam*2/np.exp(1))**(-2*nstar)
		if self.avg:
			return ans
		else:
			# If explicit, simply unwind cycle averaging.
			# This does not account for changes in sub-cycle structure with adiabaticity, but gives right average behavior.
			return ans/CycleAveragingFactor(self.Uion,E)

class YI(HybridTheories):
	"""Yudin-Ivanov phase dependent ionization rate, Eq. 17., assuming m=0."""
	def Rate(self,amp,phase):
		"""Get the phase dependent rate.  The Phase can be extracted using the
		Ionization class method ExtractEikonalForm.

		:param numpy.array amp: Amplitude of the field, any shape
		:param numpy.array phase: Phase of the field, any shape"""
		amp = amp + self.cutoff_field
		theta = (phase - 0.5*np.pi)%np.pi - 0.5*np.pi
		w = self.w0
		lstar = self.lstar
		l = self.l
		nstar,F,gam,g,nu = self.Params(amp,w,self.Uion,self.Z)
		kappa = np.log(gam+np.sqrt(gam**2+1)) - gam/np.sqrt(1+gam**2)
		pre = self.Uion
		pre *= 2**(2*nstar) / (nstar*spsf.gamma(nstar+lstar+1)*spsf.gamma(nstar-lstar)) # A_nl = |C_PPT|^2
		pre *= (2*l + 1) * np.sqrt(3*kappa/gam**3) # here 2*l+1 = B_lm with m=0
		pre *= (1+gam**2)**0.75 * (4/np.sqrt(3*np.pi)) * (gam**2/(1+gam**2)) * self.FourierSum(gam,nu) # C
		pre *= (2/F)**(2*nstar-1)
		a = 1+gam*gam-np.sin(theta)**2
		b = np.sqrt(a*a+4*gam*gam*np.sin(theta)**2)
		c = np.sqrt((np.sqrt((b+a)/2)+gam)**2 + (np.sqrt((b-a)/2)+np.sin(np.abs(theta)))**2)
		Phi = (gam**2 + np.sin(theta)**2 + 0.5)*np.log(c)
		Phi -= 3*(np.sqrt(b-a)/(2*np.sqrt(2)))*np.sin(np.abs(theta))
		Phi -= (np.sqrt(b+a)/(2*np.sqrt(2)))*gam
		return pre * np.exp(-amp**2 * Phi / w**3)
